// GBAMP Deskinner.cpp
//
/*
GBAMP Deskinner.exe
Written by chishm aka Michael Chisholm.
Dumps a BACKIMG file used by a GBA Movie Player
into its corresponding sources
*/

/*
Source Code Information, Licensing and Copyright:
I hereby release this program and all source code into
the public domain. If you wish to port the code or use
it in your own applications this is completely okay.
Just email me first at chishm@hotmail.com as a simple courtesy.
This has been written using MS VC++ version 6.0.
It does not use windows APIs for ini access. All code should be
cross platform compatible.
*/

/* 
2005-04-20 - Version 1.0
* Ini file now implemented
* First official release
* Compatible only with firmware that includes game function
*/

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string>
using namespace std;

#define PALETTE_LENGTH 1024
#define HEAD_LENGTH 54
#define backimgName "BACKIMG"
#define NUM_BMPS 12
#define NUM_OFFSETS 7
#define NUM_TEXT_COLORS 4

unsigned char BMP_HEAD[54] = {
	0x42, 0x4D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, 0x04, 0x00, 0x00, 0x28, 0x00, 
		0x00, 0x00, 0xF0, 0x00, 0x00, 0x00, 0xA0, 0x00, 0x00, 0x00, 0x01, 0x00, 0x08, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0x96, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

char* ini_BmpData[NUM_BMPS] = {
	"\"Main_BG.bmp\" , 0, 0, 240, 160",
		"\"Main_Icon_Movie.bmp\" , 0, 0, 64, 64",
		"\"Main_Icon_Music.bmp\" , 0, 0, 64, 64",
		"\"Main_Icon_Image.bmp\" , 0, 0, 64, 64",
		"\"Main_Icon_Game.bmp\" , 0, 0, 64, 64",
		"\"Main_Icon_Ebook.bmp\" , 0, 0, 64, 64",
		"\"Main_Icon_Help.bmp\" , 0, 0, 64 ,64",
		"\"File_BG.bmp\" , 0 ,0,240, 160",
		"\"File_Icon_Folder.bmp\" , 0, 0, 16, 16",
		"\"File_Icon_Previous.bmp\" , 0, 0, 16, 16",
		"\"Text_BG.bmp\" , 0, 0, 240, 160",
		"\"Music_BG.bmp\" , 0, 0, 240, 160",
};

char* ini_BmpNames[NUM_BMPS] = {
	"Main_BG","Main_Icon_Movie","Main_Icon_Music",
		"Main_Icon_Image","Main_Icon_Game","Main_Icon_EBook",
		"Main_Icon_Help","File_BG","File_Icon_Folder","File_Icon_Previous",
		"Text_BG","Music_BG"
};

char* ini_TextNames[NUM_TEXT_COLORS] = {"Menu", "Highlight", "Text", "Reserved"};

char* ini_IconNames[NUM_OFFSETS] = {
	"Movie", "Music", "Image", "Game",
		"Ebook", "Help", "FileText"
};



long offset_Pal[4] = { 0x00000, 0x0EA00, 0x18600, 0x22000 };

long offset_Img[NUM_BMPS] = { 
	0x00400, 0x09A00, 0x0AA00, 0x0BA00, 0x0CA00, 0x0DA00, 0x2BA00,
		0x0EE00, 0x18400, 0x18500, 0x18A00, 0x22400
};

long offset_textPal = 0x2CA00;

char backimgIniName[12] = {"BackImg.ini"};


typedef struct							/* Info about location of images */
{
	string		filename;				/* Name of bitmap */
	int			xOffset;				/* Offset of image data from left of bitmap */
	int			yOffset;				/* " top of bitmap */
	int			width;					/* Width of image */
	int			height;					/* Height of image */
} INI_BMP;

typedef struct							/* Positions of icons on GBAMP */
{
	unsigned char	xOffset;			/* Offset from left of screen */
	unsigned char	yOffset;			/* Offset from top of screen */
} INI_OFFSET;

typedef struct							/* Color in RGBN format */
{
	unsigned char red;
	unsigned char green;
	unsigned char blue;
	unsigned char reserved;
} RGBN;


INI_BMP iniBmpInfoFromString ( char* inString )
{
	INI_BMP bmpInfo;
	string temp1 = inString;
	
	// Goes through and extracts data from each comma delimited section in value
	bmpInfo.filename = temp1.substr(0,temp1.find_first_of(","));
	bmpInfo.filename.erase(0,bmpInfo.filename.find_first_of("\"") + 1);
	bmpInfo.filename.erase(bmpInfo.filename.find_last_of("\""));
	
	temp1.erase(0,temp1.find_first_of(",") + 1);
	bmpInfo.xOffset=atoi(temp1.c_str());
	
	temp1.erase(0,temp1.find_first_of(",") + 1);
	bmpInfo.yOffset=atoi(temp1.c_str());
	
	temp1.erase(0,temp1.find_first_of(",") + 1);
	bmpInfo.width=atoi(temp1.c_str());
	
	temp1.erase(0,temp1.find_first_of(",") + 1);
	bmpInfo.height=atoi(temp1.c_str());
	
	return (bmpInfo);
}

int writeBMP (const char* filename, char* palette, char* data, long width , long height)
{
	int i;
	FILE* bmpFile;
	bmpFile = fopen(filename,"wb");
	if (bmpFile == NULL){
		return (2);
	} else {
		
		fwrite(&BMP_HEAD, sizeof(BMP_HEAD), 1, bmpFile);
		fseek(bmpFile, 18, SEEK_SET);
		fwrite(&width, sizeof(long), 1, bmpFile);
		fwrite(&height, sizeof(long), 1, bmpFile);
		fseek(bmpFile, HEAD_LENGTH, SEEK_SET);
		fwrite(palette,PALETTE_LENGTH,1,bmpFile);
		for (i=0; i < height; i++){
			fwrite( &data[(height-i-1)*width] , width, 1, bmpFile);
		}
		fclose(bmpFile);
		return (0);
	}
}

char* extractBmpData (FILE* backimgFile, long start, long length)
{
	char* bmpData = (char*) malloc(length);
	fseek(backimgFile, start, SEEK_SET);
	fread(bmpData, length, 1, backimgFile);
	
	return bmpData;
}

char* extractBmpPal (FILE* backimgFile, long start)
{
	char* bmpPal = (char*) malloc(PALETTE_LENGTH);
	fseek(backimgFile, start, SEEK_SET);
	fread(bmpPal, PALETTE_LENGTH, 1, backimgFile);
	return bmpPal;
}


int main(int argc, char* argv[])
{
	char* bmpPal[4];
	char* bmpData[NUM_BMPS];
	INI_BMP bmpInfo[NUM_BMPS];
	INI_OFFSET iniOffsets[7];
	RGBN iniColors[4];
	unsigned short showHelp;
	int i;
	int errorCode = 0, returnCode = 0;
	long backimgFileSize;
	
	FILE* backimgFile;
	FILE* backimgIniFile;

	// Startup information
	printf("GBAMP Deskinner. Written by Chishm. Version: 1.00.\n");
	
	// Open BACKIMG file
	backimgFile = fopen (backimgName, "rb");
	if (backimgFile == NULL) {
		printf("Unable to open %s.\n",backimgName);
		errorCode = (errorCode < 2 ? 2 : errorCode);
	} else {
		  // obtain file size, display message if wrong, continue anyway
		fseek (backimgFile , 0 , SEEK_END);
		backimgFileSize = ftell (backimgFile);
		rewind (backimgFile);
		if (backimgFileSize != 182816){
			printf ("%s is wrong size.\n",backimgName);
			errorCode = (errorCode < 1 ? 1 : errorCode);
		}

		// Get all the palette information
		for (i = 0; i < 4; i++){
			bmpPal[i] = extractBmpPal(backimgFile, offset_Pal[i]);
		}
		
		// Get all the image information
		for (i = 0; i < NUM_BMPS; i++){
			bmpInfo[i] = iniBmpInfoFromString(ini_BmpData[i]);
			bmpData[i] = extractBmpData(backimgFile, offset_Img[i], bmpInfo[i].width * bmpInfo[i].height);
		}
		
		// Get the colours for the text
		fseek(backimgFile, offset_textPal, SEEK_SET);
		fread (iniColors, sizeof(RGBN), NUM_TEXT_COLORS, backimgFile);
		
		// Get the icon offsets
		fread (iniOffsets, sizeof(INI_OFFSET), NUM_OFFSETS, backimgFile);
		
		// Determine whether help text is visible
		fread (&showHelp, sizeof(unsigned short), 1, backimgFile);
		
		fclose(backimgFile);
		
		
		// Write out bitmaps for main screen
		for (i=0; i < 7; i++) {
			printf ("Writing %s.\n", bmpInfo[i].filename.c_str());
			returnCode = writeBMP(bmpInfo[i].filename.c_str(), bmpPal[0], bmpData[i], bmpInfo[i].width, bmpInfo[i].height);
			errorCode = (errorCode < returnCode ? returnCode : errorCode);
		}
		
		// Write bitmaps for file screen
		for (i=7; i < 10; i++) {
			printf ("Writing %s.\n", bmpInfo[i].filename.c_str());
			returnCode = writeBMP(bmpInfo[i].filename.c_str(), bmpPal[1], bmpData[i], bmpInfo[i].width, bmpInfo[i].height);
			errorCode = (errorCode < returnCode ? returnCode : errorCode);
		}
		
		// Write bitmap for text screen
		i = 10;
		printf ("Writing %s.\n", bmpInfo[i].filename.c_str());
		returnCode = writeBMP(bmpInfo[i].filename.c_str(), bmpPal[2], bmpData[i], bmpInfo[i].width, bmpInfo[i].height);
		errorCode = (errorCode < returnCode ? returnCode : errorCode);
		
		// Write bitmap for music screen
		i = 11;
		printf ("Writing %s.\n", bmpInfo[i].filename.c_str());
		returnCode = writeBMP(bmpInfo[i].filename.c_str(), bmpPal[3], bmpData[i], bmpInfo[i].width, bmpInfo[i].height);
		errorCode = (errorCode < returnCode ? returnCode : errorCode);
		
		// Free bitmap memory used
		for (i = 0; i < 4; i++){
			free (bmpPal[i]);
		}
		
		for (i =0; i < NUM_BMPS; i++){
			free (bmpData[i]);
		}
		
	}
	
	// Write ini file
	backimgIniFile = fopen (backimgIniName, "w");
	if (backimgIniFile == NULL) {
		printf ("Unable to open %s.\n", backimgIniName);
		errorCode = (errorCode < 2 ? 2 : errorCode);
	} else {
		printf ("Writing %s.\n", backimgIniName);

		fprintf (backimgIniFile, "[FileVersion]\n");
		fprintf (backimgIniFile, "IniVersion = 1\n");
		fprintf (backimgIniFile, "FirmwareVersion = e17\n\n");
		fprintf (backimgIniFile, "[Bitmaps]\n");
		for (i = 0; i < NUM_BMPS; i++){
			fprintf(backimgIniFile, "%s = %s\n", ini_BmpNames[i], ini_BmpData[i]);
		}
		
		fprintf (backimgIniFile, "\n[Color]\n");
		for (i = 0; i < NUM_TEXT_COLORS; i++){
			fprintf (backimgIniFile, "%s = %d,%d,%d\n",ini_TextNames[i], iniColors[i].red, iniColors[i].green, iniColors[i].blue);
		}
		
		fprintf (backimgIniFile, "\n[MenuIconPostion]\n");
		for (i = 0; i < NUM_OFFSETS; i++){
			fprintf (backimgIniFile, "%s = %d,%d\n", ini_IconNames[i], iniOffsets[i].xOffset, iniOffsets[i].yOffset);
		}
		
		fprintf (backimgIniFile, "\n[Flags]\n");
		fprintf (backimgIniFile, "ShowHelpText = %s", ( showHelp == 0 ? "False" : "True"));
		
		fclose(backimgIniFile);
		
	}
	
	switch (errorCode) {
	case 0:
		printf ("Completed successfully.\n");
		break;
	case 1:
		printf ("There was a recoverable error.\n");
		break;
	case 2:
		printf ("There was an unrecoverable error.\nOne or more files were not accessable.\n");
		break;
	default:
		printf ("An unknown error occured.\n");
	}
	
	return errorCode;
}
